/*-- scss:functions --*/
@function color-contrast($color) {
  @return if(
    red($color) * 0.299 + green($color) * 0.587 + blue($color) * 0.114 > 186,
    black, white
  );
}

/*-- scss:defaults --*/
$body-color:            color-contrast($body-bg) !default;

/*-- scss:mixins --*/
@mixin body-color {
  color: $body-color
}

/*-- scss:rules --*/
body {
  @include body-color;
}
